import os
import sys

from yaspin import yaspin
from yaspin.spinners import Spinners
from pytubefix import YouTube
from pytubefix.cli import on_progress
from termcolor import colored
from moviepy.video.io.ffmpeg_tools import ffmpeg_merge_video_audio

from pyutube.utils import (
    console,
    error_console,
    ask_resolution,
    CANCEL_PREFIX
)


class VideoService:
    def __init__(self, url: str, quality: str, path: str) -> None:
        self.url = url
        self.quality = quality
        self.path = path

    # Helper functions for the Downloader class

    def search_process(self) -> YouTube:
        """
        Performs the video search process.

        Returns:
            YouTube: An instance of the YouTube class representing the searched video.
        """
        try:
            video = self.__video_search()
        except Exception as error:
            error_console.print(f"Error: {error}")
            sys.exit(1)

        if not video:
            error_console.print("No stream available for the url.")
            sys.exit()
        return video

    @yaspin(
        text=colored("Searching for the video", "green"),
        color="green", spinner=Spinners.point
    )
    def __video_search(self) -> YouTube:
        return YouTube(
            self.url,
            use_oauth=True,
            allow_oauth_cache=True,
            on_progress_callback=on_progress,
        )

    @yaspin(
        text=colored("getting video streams", "green"),
        spinner=Spinners.point
    )
    def get_available_resolutions(self, video: YouTube) -> set:
        """
        Get a set of all available resolutions for the video.

        Args:
            video: The video to retrieve available resolutions from.

        Returns:
            set: A set containing all available resolutions.
        """
        streams = video.streams

        available_streams = streams.filter(
            progressive=False, adaptive=True, mime_type="video/mp4")
        audio_stream = streams.filter(
            only_audio=True).order_by('mime_type').first()

        resolutions_with_sizes = self.get_video_resolutions_sizes(
            available_streams, audio_stream
        )

        resolutions_with_sizes = sorted(
            resolutions_with_sizes, key=lambda x: int(
                x[0][:-1]) if x[0][:-1].isdigit() else float('inf')
        )

        resolutions, sizes = zip(*resolutions_with_sizes)
        resolutions = list(resolutions)
        sizes = list(sizes)

        return resolutions, sizes, available_streams, audio_stream

    @yaspin(
        text=colored("Downloading the video...", "green"),
        color="green", spinner=Spinners.dots13
    )
    def get_video_streams(self, quality: str, streams: YouTube.streams) -> YouTube:
        """
        Downloads the video streams based on the specified quality.
        If the specified quality is not available, it selects the nearest quality.

        Args:
            video: The video to retrieve streams from.
            quality: The desired quality of the video streams.

        Returns:
            The video stream with the specified quality,
            or the best available stream if no match is found.
        """
        stream = streams.filter(res=quality).first()

        if quality.startswith(CANCEL_PREFIX):
            error_console.print("❗ Cancel the download...")
            sys.exit()

        if not stream:
            available_qualities = [stream.resolution for stream in streams]

            available_qualities = [int(res.replace("p", "")) for res in available_qualities]

            quality_int = int(quality.replace("p", "")) if "p" in quality else int(quality)
            selected_quality = min(available_qualities,
                                   key=lambda x: abs(quality_int - x)
                                   )

            stream = streams.filter(res=str(selected_quality) + "p").first()

        return stream

    def get_selected_stream(self, video, is_audio: bool = False):
        """
        Get the selected video stream based on user preference.

        Returns:
            YouTube: The selected video stream.
        """
        resolutions, sizes,  streams, video_audio = self.get_available_resolutions(
            video)

        if not streams:
            error_console.print("❗ Cancel the download...")
            sys.exit()

        if not is_audio:
            self.quality = self.quality or ask_resolution(resolutions, sizes)

        if not self.quality and not is_audio:
            error_console.print("❗ Cancel the download...")
            sys.exit()

        # TODO:
        # [] if self.quality.startswith(CANCEL_PREFIX) else
        # return streams, video_audio

        return streams, video_audio, self.quality

    def merging(self, video_name: str, audio_name: str):
        """
        Merges the video and audio files into a single file.

        Args:
            video_name: The name of the video file.
            audio_name: The name of the audio file.

        Returns:
            None
        """
        output_directory = os.path.join(self.path, "output")
        os.makedirs(output_directory, exist_ok=True)

        # Output file path
        base_name = os.path.splitext(os.path.basename(video_name))[0]
        output_file = os.path.join(output_directory, f"{base_name}.mp4")
        # output_file = os.path.join(output_directory, os.path.basename(
        #     video_name).replace(os.path.splitext(video_name)[1], '.mp4'))

        # Extract base names to match files
        video_base_name = os.path.splitext(os.path.basename(video_name))[0]
        audio_base_name = os.path.splitext(os.path.basename(audio_name))[0]

        # Locate the video file
        video_path = None
        for file in os.listdir(self.path):
            if file.startswith(video_base_name):
                video_path = os.path.join(self.path, file)
                break

        if video_path is None:
            error_console.print(f"❗ Video file not found: {video_name}")
            sys.exit(1)

        # Locate the audio file
        audio_path = None
        for file in os.listdir(self.path):
            if file.startswith(audio_base_name):
                audio_path = os.path.join(self.path, file)
                break

        if audio_path is None:
            error_console.print(f"❗ Audio file not found: {audio_name}")
            sys.exit(1)

        try:
            # Merge video and audio
            ffmpeg_merge_video_audio(
                video_path,
                audio_path,
                output_file,
                # vcodec='copy',
                # acodec='copy',
                logger=None
            )

            # Remove original files
            os.remove(video_path)
            os.remove(audio_path)

            # Move the merged file to the current directory
            if os.path.exists(output_file):
                merged_file_name = os.path.basename(output_file)
                parent_directory = os.path.dirname(output_directory)

                final_path = os.path.join(parent_directory, merged_file_name)

                os.replace(output_file, final_path)
                os.rmdir(output_directory)
            else:
                error_console.print("❗ Merged video file not found in the output directory.")
                sys.exit(1)

        except Exception as e:
            error_console.print(f"❗ An error occurred: {str(e)}")
            sys.exit(1)

    @staticmethod
    def get_video_resolutions_sizes(
            available_streams: list[YouTube],
            audio_stream: YouTube
    ) -> list:
        """
        Get the available video resolutions.

        Args:
            available_streams: The available video streams.
            audio_stream: The audio stream.

        Returns:
            list: The available video resolutions.
        """
        if not available_streams:
            return []

        # Calculate the total audio file size in bytes
        audio_filesize = audio_stream.filesize

        resolutions_with_sizes = []
        one_mb = 1024 * 1024
        one_gb = one_mb * 1024
        for stream in available_streams:
            if stream.resolution:
                # Calculate the total video file size including audio in bytes
                video_filesize_bytes = stream.filesize

                if not stream.is_progressive:
                    video_filesize_bytes += audio_filesize

                # Convert the video file size to KB or MB dynamically
                if video_filesize_bytes >= one_gb:
                    # If size is >= 1 GB
                    video_filesize = \
                        f"{video_filesize_bytes / (one_gb):.4f} GB"

                elif video_filesize_bytes >= one_mb:
                    # If size is >= 1 MB
                    video_filesize = \
                        f"{video_filesize_bytes / (one_mb):.2f} MB"
                else:
                    video_filesize = f"{video_filesize_bytes / 1024:.2f} KB"

                resolutions_with_sizes.append(
                    (stream.resolution, video_filesize))

        return resolutions_with_sizes



"""This is the initialization module for the pyutube package.

It provides the main entry point for the package by importing the `app`
object from the `cli` submodule.

Example:
    To use this package, you can import the `app` object directly:

    >>> from pyutube import app
    >>> app.run()

"""
from .cli import app


# List the symbols you want to export from this module
# __all__ = ['app']



from pyutube.handlers.PlaylistHandler import PlaylistHandler
import os
import sys

from pytubefix import YouTube
from pytubefix.helpers import safe_filename

from pyutube.utils import asking_video_or_audio, console, error_console
from pyutube.services.AudioService import AudioService
from pyutube.services.VideoService import VideoService
from pyutube.services.FileService import FileService


class DownloadService:
    def __init__(
            self, url: str, path: str, quality: str, is_audio: bool = False, make_playlist_in_order: bool = False,
    ):
        self.url = url
        self.path = path
        self.quality = quality
        self.is_audio = is_audio

        self.make_playlist_in_order = make_playlist_in_order

        self.video_service = VideoService(self.url, self.quality, self.path)
        self.audio_service = AudioService(url)
        self.file_service = FileService()

    def download(self, title_number: int = 0) -> bool:
        video, video_id,  streams, video_audio, self.quality = self.download_preparing()

        if self.is_audio:
            self.download_audio(video, video_audio, video_id, title_number)
        else:
            video_file = self.video_service.get_video_streams(self.quality, streams)
            if not video_file:
                error_console.print(
                    "Something went wrong while downloading the video.")
                sys.exit()

            return self.download_video(video, video_id, video_file, video_audio, title_number)

        return True

    def download_audio(self, video: YouTube, video_audio: YouTube, video_id: str, title_number: int = 0) -> bool:
        audio_filename = self.file_service.generate_filename(video_audio, video_id, is_audio=True)

        if self.make_playlist_in_order:
            base_name, extension = os.path.splitext(audio_filename)
            audio_filename = f"{title_number}__{base_name}{extension}"

        audio_filename = self.file_service.handle_existing_file(
            video, video_id, audio_filename, self.path, self.is_audio)

        try:
            if self.is_audio:
                console.print("⏳ Downloading the audio...", style="info")

            self.file_service.save_file(video_audio, audio_filename,  self.path)

        except Exception as error:
            error_console.print(
                f"❗ Error (please report this in github issue: https://github.com/Hetari/pyutube/issues):\n {error}")
            sys.exit()

        if self.is_audio:
            console.print("\n\n✅ Download completed", style="success")
            return True
        return audio_filename

    def download_video(self, video: YouTube, video_id: str, video_stream: YouTube, video_audio: YouTube,
                       title_number: int = 0) -> bool:

        # Generate filename with title, quality, and file extension
        video_filename = self.file_service.generate_filename(video_stream, video_id)

        # Prepend the title number and `__` to the filename if ordering is required
        if self.make_playlist_in_order:
            video_base_name, video_extension = os.path.splitext(video_filename)
            video_filename = f"{title_number}__{video_base_name}{video_extension}"

        # Handle existing files
        video_filename = self.file_service.handle_existing_file(
            video, video_id, video_filename, self.path, self.is_audio)

        try:
            console.print("⏳ Downloading the video...", style="info")

            self.file_service.save_file(video_stream, video_filename, self.path)
            audio_filename = self.download_audio(video, video_audio, video_id, title_number)

            video_base_name, video_extension = os.path.splitext(video_filename)
            audio_base_name, audio_extension = os.path.splitext(audio_filename)
            video_safe_filename = f"{safe_filename(video_base_name)}{video_extension}"
            audio_safe_filename = f"{safe_filename(audio_base_name)}{audio_extension}"

            self.video_service.merging(video_safe_filename, audio_safe_filename)

        except Exception as error:
            error_console.print(
                f"❗ Error (please report this in github issue: https://github.com/Hetari/pyutube/issues):\n {error}")
            sys.exit()

        console.print("\n\n✅ Download completed", style="success")
        return self.quality

    def asking_video_or_audio(self):
        try:
            self.is_audio = asking_video_or_audio()
        except TypeError:
            return
        self.download()

    def get_playlist_links(self):
        handler = PlaylistHandler(self.url, self.path)
        new_path, is_audio, videos_selected, make_in_order, playlist_videos = handler.process_playlist()
        self.make_playlist_in_order = make_in_order
        selected_titles = [
            title for title, video_id in playlist_videos if video_id in videos_selected
        ]

        # Download the selected videos
        for index, video_id in enumerate(videos_selected):
            url = f"https://www.youtube.com/watch?v={video_id}"
            if make_in_order:
                i = int(selected_titles[index].split('__')[0])
            else:
                i = ''

            self.url = url
            self.path = new_path
            self.is_audio = is_audio

            if index == 0:
                # If it is the first video, download it and store the quality
                self.video_service = VideoService(self.url, self.quality, self.path)
                quality = self.download(i)
                continue

            # If it is not the first video, download it with the stored quality
            self.quality = quality
            self.video_service = VideoService(self.url, self.quality, self.path)
            self.download(i)

    def download_preparing(self):
        video = self.video_service.search_process()
        console.print(f"Title: {video.title}\n", style="info")
        video_id = video.video_id
        streams, video_audio, self.quality = self.video_service.get_selected_stream(video, self.is_audio)

        return video, video_id,  streams, video_audio, self.quality



from .DownloadService import DownloadService
from .VideoService import VideoService
from .AudioService import AudioService
from .FileService import FileService


__all__ = ['DownloadService', 'VideoService', 'AudioService', 'FileService']



"""
This module contains the utils functions for the pyutube package.
"""

import subprocess
import sys
import os

import requests
import inquirer
from yaspin import yaspin
from yaspin.spinners import Spinners
from rich.console import Console
from rich.theme import Theme
from termcolor import colored
from pytubefix import __version__ as pytubefix_version


__version__ = "1.4.16"
__app__ = "pyutube"
ABORTED_PREFIX = "Aborted"
CANCEL_PREFIX = "Cancel"


# Set up the console
custom_theme = Theme({
    "info": "#64b0f2",
    "warning": "color(3)",
    "success": "green",
})
console = Console(theme=custom_theme)
error_console = Console(stderr=True, style="red")


def clear() -> None:
    """
    Function to clear the console screen, it can be used for any operating system

    Args:
        This function does not take any parameters.

    Returns:
        It does not return anything (None).
    """
    # For Windows
    if os.name == "nt":
        os.system("cls")
    else:
        # For Unix/Linux/MacOS
        os.system("clear")


@yaspin(text="Checking internet connection", color="blue", spinner=Spinners.earth)
def is_internet_available() -> bool:
    """
    Checks if internet connection is available by making a simple request
    to http://www.google.com with a timeout of 5 seconds.

    Returns:
        bool: the request status (True if available, False if not).
    """
    try:
        requests.get("https://www.google.com", timeout=5)
        return True
    except Exception:
        return False


def file_type() -> str:
    """
    Prompts the user to choose a file type for download and returns
    the chosen file type as a string.

    Args:
        None

    Returns:
        str: The chosen file type as a string.
    """
    # make the console font to red
    questions = [
        inquirer.List(
            "file_type",
            message="Choose the file type you want to download",
            choices=['Audio', 'Video', CANCEL_PREFIX],
        ),
    ]

    try:
        answer = inquirer.prompt(questions)["file_type"]

    # TypeError: 'NoneType' object is not subscriptable

    except TypeError:
        return ABORTED_PREFIX

    except Exception as error:
        error_console.print(f"Error: {error}")
        sys.exit()

    return answer


def ask_resolution(resolutions: set, sizes) -> str:
    """
    Prompts the user to choose a resolution for download
    and returns the chosen resolution as a string.

    Args:
        resolutions (set): The set of available resolutions.

    Returns:
        str: The chosen resolution as a string.
    """
    # Create a dictionary to relate each size with its resolution
    size_resolution_mapping = dict(zip(resolutions, sizes))

    # Generate the choices for the user prompt
    resolution_choices = [
        f"{size} ~= {resolution}" for size, resolution in size_resolution_mapping.items()
    ] + [CANCEL_PREFIX]

    questions = [
        inquirer.List(
            "resolution",
            message="Choose the resolution you want to download",
            choices=resolution_choices,
        ),
    ]

    try:
        answer = inquirer.prompt(questions)["resolution"]

    except TypeError:
        return ABORTED_PREFIX

    except Exception as error:
        error_console.print(f"Error: {error}")
        sys.exit()

     # Extract the resolution part from the user's choice
    return answer.split(" ~= ")[0]


def ask_rename_file(filename: str) -> str:
    """
    Function to ask the user whether to rename, overwrite, or cancel the file operation.

    Args:
        filename (str): The name of the file.

    Returns:
        str: The user's choice to rename, overwrite, or cancel the file operation.
    """
    console.print(
        f"'{filename}' is already exists, do you want to:", style="info")
    questions = [
        inquirer.List(
            "rename",
            message="Do you want to",
            choices=['Rename it', 'Overwrite it', CANCEL_PREFIX.capitalize()],
        ),
    ]
    return inquirer.prompt(questions)["rename"]


def ask_playlist_video_names(videos):
    note = colored("NOTE:", "cyan")
    select_one = colored("<space>", "red")
    select_all = colored("<ctrl+a>", "red")
    invert_selection = colored("<ctrl+i>", "red")
    restart_selection = colored("<ctrl+r>", "red")

    print((
        f"{note} Press {select_one} to select the videos, {select_all} to select all, "
        f"{invert_selection} to invert selection, and {restart_selection} to restart selection"
    ))

    questions = [
        inquirer.Checkbox(
            "names",
            message="Choose the videos you want to download",
            choices=videos,
        ),
    ]

    try:
        answer = inquirer.prompt(questions)["names"]

    except TypeError:
        return ABORTED_PREFIX

    except Exception as error:
        error_console.print(f"Error: {error}")
        sys.exit()

    return answer


def ask_for_make_playlist_in_order():
    # make_in_order = colored( "", "cyan")

    questions = [
        inquirer.Confirm(
            "ask_for_make_playlist_in_order",
            message="Do you want the playlist videos in order? ",
            default=False
        ),
    ]

    try:
        answer = inquirer.prompt(questions)["ask_for_make_playlist_in_order"]

    except TypeError:
        return ABORTED_PREFIX

    except Exception as error:
        error_console.print(f"Error: {error}")
        sys.exit()

    return answer


def check_for_updates() -> None:
    """
    A function to check for updates of a given package or packages.

    Returns:
        None
    """
    libraries = {
        'pyutube': {
            'version': __version__,
            'repository': 'https://github.com/Hetari/pyutube'
        },
        'pytubefix': {
            'version': pytubefix_version,
            'repository': 'https://github.com/Hetari/pytubefix'
        }
    }

    try:
        for library, version in libraries.items():
            r = requests.get(
                f'https://pypi.org/pypi/{library}/json', headers={'Accept': 'application/json'})
            if r.status_code == 200:
                latest_version = r.json()['info']['version']

                if latest_version != version['version']:
                    console.print(
                        f"👉 A new version of [blue]{library}[/blue] is available: {latest_version} " +
                        f"Updating it now... ",
                        style="warning"
                    )
                    # auto-update the package
                    try:
                        subprocess.check_call(
                            [sys.executable, '-m', 'pip', 'install', '--upgrade', library],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE
                        )
                        console.print(
                            f"✅ Successfully updated [blue]{library}[/blue] to version {latest_version}.",
                            style="success"
                        )
                    except subprocess.CalledProcessError as e:
                        error_console.print(
                            f"❗ Failed to update [blue]{library}[/blue]: {e.stderr.decode()}"
                        )
                        console.print(
                            f"❗ If you want to use the latest version of [blue]{library}[/blue], " +
                            "Update it by running [bold red link=https://github.com/Hetari/pyutube] " +
                            f"pip install --upgrade {library}[/bold red link]"
                        )

            else:
                error_console.print(
                    f"❗ Error checking for updates: {r.status_code}")
    except Exception as error:
        error_console.print(f"❗ Error checking for updates: {error}")


# main utils
def check_internet_connection() -> bool:
    """
    Checks if an internet connection is available.

    Returns:
        bool: True if internet connection is available, False otherwise.
    """
    if not is_internet_available():
        error_console.print("❗ No internet connection")
        return False

    console.print("✅ There is internet connection", style="success")
    console.print()
    return True


def asking_video_or_audio() -> bool:
    """
    Handles video link scenario.

    Args:
        None

    Returns:
        bool: True if the video link is valid, False otherwise.
    """
    file_type_choice = file_type().lower()
    is_audio = file_type_choice.startswith("audio")

    if file_type_choice.startswith(CANCEL_PREFIX.lower()):
        error_console.print("❗ Cancel the download...")
        sys.exit()

    return is_audio
